# ExclusionA_V3.py
"""========================================================================
Intersection of two plane Bézier curves 

Step 1 :
    an estimate of the intersection points are determined according 
    - the exclusion principle
    - and the subdivision principle
    with a recursive implementation
    This step provides initial values for Step 2

Step 2 :
    the intersection points are evaluated with the Newton-Raphson method
    in the parameter domain

The recursive function in Step 1 stops :
- when subdivided Bezier curves no more intersect in the 2D plane, 
- or when the lengths d1 and d2 of the 2 intervals containing 
a root-intersection are smaller than a given epsilon value

Due to recursivity, we possibly get several intervals 
containing the same root
==> we regroup them in the parametric case with the function UpdateList()
========================================================================"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.special import binom

#######################
#  BEZIER FUNCTIONS
#######################

def Bernstein(n, k, x):
    coeff = binom(n, k)
    return coeff * x**k * (1 - x)**(n - k)

def BezierPlot(ax,xptcontrole,yptcontrole,color3):
    """ Display the whole Bézier curve
    """
    degree = len(xptcontrole) - 1
    nbti = 200
    ttrace = np.linspace(0,1,nbti)
    tx = np.zeros(nbti)
    ty = np.zeros(nbti)
    for k in range(degree+1):
        poly = Bernstein(degree,k,ttrace)
        tx = (tx + xptcontrole[k]*poly)
        ty = (ty + yptcontrole[k]*poly)
    ax.plot(tx,ty,color3,lw=1)
    plt.draw()

def BezierPlotOnePoint(ax,xptcontrole,yptcontrole,t0,color7):
    """ Display value P(t0) of the Bézier curve
    """
    degree = len(xptcontrole) - 1
    x0 = 0.
    y0 = 0.
    for k in range(degree+1):
        poly = Bernstein(degree,k,t0)
        x0 = x0 + xptcontrole[k]*poly
        y0 = y0 + yptcontrole[k]*poly
    ax.plot(x0,y0,color7,ms=8)
    plt.draw()

def PolygonAcquisition(ax,color1,color2) :
    """ Acquisition of a 2D polygon in the window with subplot "ax" 
        right click stop the acquisition
    """
    x = []  # x is an empty list
    y = []
    coord = 0
    while coord != []:
        coord = plt.ginput(1, mouse_add=1, mouse_stop=3, mouse_pop=2)
        # coord is a list of tuples : coord = [(x,y)]
        if coord != []:
            xx = coord[0][0]
            yy = coord[0][1]
            ax.plot(xx,yy,color1,ms=8)
            x.append(xx)
            y.append(yy)
            plt.draw()
            if len(x) > 1 :
                ax.plot([x[-2],x[-1]],[y[-2],y[-1]],color2)
    return x,y

def OneSubdivBezierCurve(xpt, ypt):
    """ One subdivision of the input control polygon according parameter 1/2
        Return the two subdivided control polygons 
            xpt1, ypt1
            xpt2, ypt2
    """
    n = len(xpt) - 1 # degree of the Bézier curve
    xp = np.zeros(2*n+1)
    yp = np.zeros(2*n+1)
    # Initialization
    for i in range(n+1):
        xp[2*i] = xpt[i]
        yp[2*i] = ypt[i]
    # Main loop
    for j in range(1,n+1):
        for i in range(n+1-j):
            k = j + 2*i
            xp[k] = ( xp[k-1] + xp[k+1] ) / 2.
            yp[k] = ( yp[k-1] + yp[k+1] ) / 2.
    # We split the subivided polygon in two control polygons
    xpt1 = xp[0:n+1]
    xpt2 = xp[n:2*n+1]
    ypt1 = yp[0:n+1]
    ypt2 = yp[n:2*n+1]  
    return xpt1, ypt1, xpt2, ypt2

#######################
# BOXES FUNCTIONS
#######################

def MinMaxBoxCP(xControlPt, yControlPt):
    """ Determination of the rectangular min-max box that contains 
        all control points
        Return the tuple Box = (Bx, By) with :
            Bx = np.array([xmin, xmax])
            By = np.array([ymin, ymax])
    """
    Bx = np.array([min(xControlPt), max(xControlPt)])
    By = np.array([min(yControlPt), max(yControlPt)])
    Box = (Bx, By)
    return Box

def IntersectBox1D(I1, I2):
    """ Intersection of the 1-dimensional boxes I1 and I2
        precisely, here I1 and I2 are intervals : 
            I1 = [a1,b1] with a1 <= b1 , I1 is an np.array
            I2 = [a2,b2] with a2 <= b2 , I2 is an np.array
        return  inter = 1 if the two intervals intersect, = 0 if not
                K = the intersection interval (np.array)
                  = np.array([]) if inter = 0
    """
    if ( I1[1] < I2[0] ) or ( I1[0] > I2[1] ) :
        K = np.array([])
        inter = 0
    else:
        K = I1
        inter = 1
        if I1[0] < I2[0] :
            K[0] = I2[0]
        if I1[1] > I2[1] :
            K[1] = I2[1]
    return inter, K

def IntersectBox2D(Box1, Box2):
    """ Intersection of the 2-dimensional boxes Box1 and Box2
        A box is a tuple Box = (Bx, By) with :
            Bx = np.array([xmin, xmax])
            By = np.array([ymin, ymax])
        Return  1 if the boxes intersect, 0 if they do not intersect
                the intersection box in case Box1 and Box2 intersect
    """
    box = ()
    interx, Kx = IntersectBox1D(Box1[0], Box2[0])
    intery, Ky = IntersectBox1D(Box1[1], Box2[1])
    inter = interx * intery
    if inter == 1 :
        box = (Kx, Ky)
    return inter, box

def BoxDiag(Box):
    """ Computation of the diagonal value of the box
    """
    d = ( Box[0][1] - Box[0][0] )**2 + ( Box[1][1] - Box[1][0] )**2
    return np.sqrt(d)

def BoxRoot(Box):
    # return the center point of the box
    rootx = (Box[0][0] + Box[0][1]) / 2.
    rooty = (Box[1][0] + Box[1][1]) / 2.
    return rootx, rooty

def PlotBox(ax, box, color4):
    """ Plot the rectangular box = (Bx, By)
            Bx = np.array([xmin, xmax])
            By = np.array([ymin, ymax])
        in the window with subplot "ax"
    """
    x1 = box[0][0]
    x2 = box[0][1]
    y1 = box[1][0]
    y2 = box[1][1]
    x = [x1, x2, x2, x1, x1]
    y = [y1, y1, y2, y2, y1]
    ax.plot(x,y,color4,lw=1)


##########################
# THE RECURSIVE FUNCTION
##########################

def BezierIntersection(ax,xp1,yp1,I1,xp2,yp2,I2,eps):
    """ Recursive intersection of the 2 Bézier curves
        xp1,yp1,xp2,yp2 : control points of each curve
        I1, I2 : associated interval for each curve (included in [0,1])
        
        this function stops when :
        - the 2 subdivided boxes intersect (inter = True) and
          len(I1)/2 < eps  and  len(I2)/2 < eps
        or when :
        - inter = False (no more intersection between the 2 boxes)
    """   
    # subdivision of each Bézier curve

    ---
    ---

    # bounding box of each subdivided control polygon
    ---
    ---
    
    # tests and recursivity ...
    ---
    ---



##############################################
# OPTIMIZATION OF THE LISTS OF INTERSECTIONS
##############################################

def UpdateList(ListInter):
    # groups and intersects intervals containing the same root
    NewListInter = []
    while ListInter != []:
        K0 = ListInter[0]
        del ListInter[0]
        k = len(ListInter) - 1   # index of last element
        while k >= 0:
            K = ListInter[k]
            # in case of contiguous intervals
            K[0] = K[0] - 1e-8
            K[1] = K[1] + 1e-8
            inter, K1 = IntersectBox1D(K0, K)
            if inter:
                K0 = K1
                del ListInter[k]
            k = k-1
        NewListInter.append(K0)
    return NewListInter


#################################################
# NEWTON FUNCTIONS
#################################################

def BezierDerivative1D(PC,t0):
    """ PC : list of scalar values (control points)
        Return value P'(t0) of the Bézier function 
    """
    degree = len(PC) - 1
    PC = np.array(PC)
    # Control points of the derivative
    DPC = PC[1:] - PC[:-1]
    # evaluation
    Pprime = 0.
    # derivative of degree - 1
    for k in range(degree):
        poly = Bernstein(degree-1,k,t0)
        Pprime = Pprime + DPC[k]*poly
    return degree * Pprime

def BezierValue1D(PC,t0):
    """ PC : list of scalar values (control points)
        Return value P(t0) of the Bézier function
    """
    degree = len(PC) - 1
    Pval = 0.
    for k in range(degree+1):
        poly = Bernstein(degree,k,t0)
        Pval = Pval + PC[k]*poly
    return Pval

def Newton2Var(xp1,yp1,I1,xp2,yp2,I2,NbSteps):
    # initial values
    u0 = (I1[0] + I1[1]) / 2.
    v0 = (I2[0] + I2[1]) / 2.
    X0 = np.array([u0,v0])
    
    for k in range(NbSteps):
        u0 = X0[0]
        v0 = X0[1]
        # Inverse D of Jacobian matrix : D = J^-1
        D = np.zeros((2,2))
        D[0][0] = - BezierDerivative1D(yp2,v0)
        D[1][0] = - BezierDerivative1D(yp1,u0)
        D[0][1] = BezierDerivative1D(xp2,v0)
        D[1][1] = BezierDerivative1D(xp1,u0)
        D = D / np.linalg.det(D)
        # Second member
        b = np.zeros(2)
        b[0] = BezierValue1D(xp1,u0) - BezierValue1D(xp2,v0)
        b[1] = BezierValue1D(yp1,u0) - BezierValue1D(yp2,v0)
        # recurrence
        X0 = X0 - np.dot(D,b)
    return X0



############################################
if __name__ == '__main__':
    
    minmax = 5
    fig = plt.figure(figsize=(8,8))
    ax = fig.add_subplot(111)
    ax.set_xlim((-minmax,minmax))
    ax.set_ylim((-minmax,minmax))
    ax.set_xlabel('x-axis')
    ax.set_ylabel('y-axis')
    ax.set_title("Acquisition window") 

    # List of intervals for each curve that contain a root-intersection
    ListInter1 = []
    ListInter2 = []

    # Bézier curve acquisition
    xp1, yp1 = PolygonAcquisition(ax,'ob','b')
    BezierPlot(ax,xp1,yp1,'b')
    xp2, yp2 = PolygonAcquisition(ax,'or','r')
    BezierPlot(ax,xp2,yp2,'r')

    # the bounding boxes
    box1 = MinMaxBoxCP(xp1, yp1)
    PlotBox(ax, box1, '--b')
    box2 = MinMaxBoxCP(xp2, yp2)
    PlotBox(ax, box2, '--r')
    
    # initialization... and go !
    eps = 0.005
    inter, box = IntersectBox2D(box1, box2)
    if inter :
        I1 = [0,1]
        I2 = [0,1]
        BezierIntersection(ax,xp1,yp1,I1,xp2,yp2,I2,eps)
    
    if ListInter1 == []:
        print('--> THE CURVES DO NOT INTERSECT')

    #----------------------------------------
    # intervals for intersections
    print("INITIAL Lists of intersections :")
    print("First list of intersections 1 = ",ListInter1)
    print("First list of intersections 2 = ",ListInter2)

    print("------- len(ListInter1) = ",len(ListInter1))
    print("------- len(ListInter2) = ",len(ListInter2))


    NewListInter1 = UpdateList(ListInter1)
    NewListInter2 = UpdateList(ListInter2)

    print("NEW Lists of intersections :")
    print("FINAL list of intersections 1 = ",NewListInter1)
    print("FINAL list of intersections 2 = ",NewListInter2)
    
    print("------- len(NewListInter1) = ",len(NewListInter1))
    print("------- len(NewListInter2) = ",len(NewListInter2))

    #----------------------------------------------
    # display of intersection points on each curve
    for i1 in range(len(NewListInter1)):
        K = NewListInter1[i1]
        t0 = (K[0]+K[1])/2.
        BezierPlotOnePoint(ax,xp1,yp1,t0,'ob')

    for i2 in range(len(NewListInter2)):
        K = NewListInter2[i2]
        t0 = (K[0]+K[1])/2.
        BezierPlotOnePoint(ax,xp2,yp2,t0,'or')

    #----------------------------------------------
    # NEWTON PROCESS :
    # a better solution is evaluated through Newton iteration
    # with mid-point of each interval as initial values
    for k in range(len(NewListInter1)):
        I1 = NewListInter1[k]
        I2 = NewListInter2[k]
        X0 = Newton2Var(xp1,yp1,I1,xp2,yp2,I2,5)
        u0 = X0[0]
        v0 = X0[1]
        # display of intersection points on each curve
        BezierPlotOnePoint(ax,xp1,yp1,u0,'+b')
        BezierPlotOnePoint(ax,xp2,yp2,v0,'xr')







